
/* ThermometerTwo, ATMEGA168/328 Version 1.0.0*/
/* This program is designated to be PUBLIC DOMAIN */
/* Author: Lewis Balentine, October 2013 */


/* This version of the prgram drops the abiltiy to store data in the EEPROM 
while enabling collecting data from up to eight sensors. 
A boolean array named "Sensors[8]" has been added to indicate which Analog
pins have sensors attached. The array is converted to Bits and stored in
EEPROM as a single Byte at location EEsensors. The Accumalator Variables has 
been converted to an Array. Ref Voltage and Offset are the same for sensors. 
ID varies by the addition of the pin number. 

This version also adds saving the operational modes to the EEPROM.
*/

#include <avr/sleep.h>                      // needed for shutdown function
#include <EEPROM.h>                         // needed for EEPROM read and write

// EEPROM Address Constants
const word EEunused =  0;                   //  2 byte location -- reserved EEMODE FALG
const word EErefvolt=  2;                   //  2 byte location of RefVoltage 
const word EEoffset =  4;                   //  2 byte location of DegreeOffset 
const word EEminutes=  6;                   //  2 byte location of Report Target Minutes
const word EEsensors=  8;                   //  1 byte location of attached sensors
const word EEopmodes=  9;                   //  1 byte location of operational modes
const word EEunused1= 10;                   //  2 byte location -- reserved -- unused
const word EEunused2= 12;                   //  2 byte location -- reserved -- unused
const word EEunused3= 14;                   //  2 byte location -- reserved -- unused
const word EEidtring= 16;                   // ID string w/o termiantion size (16) 
const word EEidsize = 16;                   // 24 byte location of IdString 
const word EEwdsize = EEidtring+EEidsize;   // Working data storage size      (32)
//-----------------------------------------------------------
const word StorageWorking=EEunused;           // EEPROM start for working calibration data
// EEPROM start for backup  copy of constants 
// note we have to add 1 to the value
// Becasue addresses begin with zero not one
const word StorageBackup =((E2END-(EEwdsize))+1);       

// Conversion Factors/Calibraton Data
float RefVoltage;                           // Analog Sensor Reference Voltage
const float CovrtFactorV= 8192;
float DegreeOffset;                         // Fudge factor to adjust output
const float CovrtFactorO= 1024;             
char  IdString[EEidsize+1];                 // ID/Location string for this device
word  MinuteTarget      = 1;                // Number of minutes between report lines
byte  newflg=0;                             // used to indicate new factors in memory

float Celsius;                              // Last conversion to Celsius Temperature
float Fahrenheit;                           // Last conversion to Fahrenheit Temperature

// Global operational mode Variables        // set default operation modes
boolean       ReportMode    = true;         // True =  reporting, False = Command Mode 
boolean       RtnRawRead    = true;         // True =  include Raw
boolean       RtnCelsius    = true;         // True =  include Celsius
boolean       RtnFahrenh    = true;         // True =  include Fahrenheit
boolean       DeBug         = false;        // True =  extended reporting for debugging
boolean       RoundMode     = true;         // True =  round output to nearest 1/4 or 1/2
boolean       EepromMode    = false;        // not used in this program --- for compatibility only
boolean       Sensors[8];                   // True = Sensor Attached, Report Sensor

// Global work Variables                    // There are 1000 milliseconds in a second
// word        SecondsMinute = 10000;       // --- to speed things up a bit for debugging
word          SecondsMinute = 60000;        // added so calibraton timining can be reduced
unsigned long SecondsTarget = 0;            // Number of seconds between report lines
unsigned long RptStartTime  = 0;            // Time between report lines 
byte          gap           = 0;            // used to increase gap between reads
unsigned long RptTrigger    = 0;            // Target Time for report 
unsigned long CycleStart    = 0;            // Target Time for report 
unsigned long CycleTime     = 0;            // Target Time for report 

unsigned long Accumalator[8]= {0,0,0,0,0,0,0,0};  // Accumalate temperature reads
unsigned long CycleCount    = 0;            // Cycles per Report line
char          cmd[]         = {0,0,0};      // used to store two character command
char          prevcmd[]     = {0,0,0};      // used to store previous two character command

// built in LED on Pin Digital Pin 13
static const uint8_t LED = 13;     // Nano Board (16)
// Analog Pin Numbers for NANO Version 3.0
// These constants match the labels on the Nano board
static const uint8_t A00 = 14;     // Nano Board (19)
static const uint8_t A01 = 15;     // Nano Board (20)
static const uint8_t A02 = 16;     // Nano Board (21)
static const uint8_t A03 = 17;     // Nano Board (22)
static const uint8_t A04 = 18;     // Nano Board (23)
static const uint8_t A05 = 19;     // Nano Board (24)
static const uint8_t A06 = 20;     // Nano Board (25)
static const uint8_t A07 = 21;     // Nano Board (26)
// Pin Number array
byte PinNo[8]= {A00,A01,A02,A03,A04,A05,A06,A07};   
                                            
//=====================================================================================
void setup() 
  { char c;
  
    Serial.begin (9600);
    pinMode(LED, OUTPUT);                    // so we can blink it later during writes
    
    EnableADC();                             // enables the ADC and set ADC clock factor
    delay (1000);                            // let serial library complete setup
    while (Serial.available()>0)             // drain any data from the serial buffer 
        c=Serial.read();
    Read_Calibration_Data();                 // read and set conversion factors from EEPROM
    // calculate seconds between report lines
    // we have to "cast" the two word values or we will get a word value for the result
    SecondsTarget=long(MinuteTarget)*long(SecondsMinute);
    //delay (5000);                          // allow PC 5 seconds to get setup
    //if (EepromMode==false) ReportStatus();   // report default parameters
    ReportStatus();
    for (byte i=0; i<8; i++) Accumalator[i] = 0;    // set startup parameters
    CycleCount  = 0;
    for (byte i=0; i<8; i++) Accumalator[i] = 0;    // set startup parameters
    RptTrigger  = millis() + SecondsTarget;
    RptStartTime= millis(); 
  }

//-----------------------------------------------------------
void loop() 
  { char c1, c2;
    word wtemp;
    
    // ---- This is where we check for command input
    if ((Serial.available()>2) && (EepromMode==false))
      { if(ReadTwoCharacters()) CmdProcessor();}
     
    // ---- This is where we collect our temperature data
    // gap is used to increase the amount of time between reading sampling the ADC.
    // This Insures that we will not miss any data transmitted on the serial port.
    if (gap++ == 9)           
      { gap=0;                     
        // cycle times are only used if debugging is turned on
        CycleStart= millis(); 
        ReadRawTemp();
        CycleTime = CycleTime+(millis()-CycleStart);
      }
    
    // ---- This is where we output the teperature data
    // The time required to read 64 samples is about 119-120 milliseconds. If we get
    // within 150 milliseconds of the Report Trigger Time then we wait for it.
    // With the these timing numbers there are 500 reads of 
    // 64 virtual 12 bit samples per minute.
    // Added condition for millis exceeding report trigger (possible with long commands)    
    if (((RptTrigger-millis())< 150) || (millis()>RptTrigger))
      { // Serial.println ("Got here: RptTrigger ");
        while (millis() < RptTrigger);
        // We want the new trigger time set as close as possible to when the previous trigger 
        // went off --- so we put ti first.
        RptTrigger= (millis() + (SecondsTarget));
        Report();
      }
  } 

//================================================================  
void CmdProcessor()
  {// this function is the main command handler
   // not many comments because I think the code is obviuos
   if (DeBug == true)
       { Serial.print (F("; Command Processor "));
         DebugPrintCharacters (cmd[0],cmd[1]);
        }
    if      ((cmd[0]=='C') && (cmd[1]=='=')) CelsiusEquals();
    else if ((cmd[0]=='D') && (cmd[1]=='B')) ToggleDebugMode();
    else if ((cmd[0]=='D') && (cmd[1]=='O')) NewDegreeOffset();
    else if ((cmd[0]=='F') && (cmd[1]=='=')) FahrenheitEquals();
    else if ((cmd[0]=='I') && (cmd[1]=='D')) {Serial.print(F("; ")); Print_IdString();}
    else if ((cmd[0]=='L') && (cmd[1]=='L')) HelpMe();
    else if ((cmd[0]=='L') && (cmd[1]==':')) NewIdString();
    else if ((cmd[0]=='R') && (cmd[1]=='V')) NewRefVolt();
    else if ((cmd[0]=='S') && (cmd[1]=='S')) ShutDown();
    else if ((cmd[0]=='S') && (cmd[1]=='T')) ReportStatus();
    else if ((cmd[0]=='W') && (cmd[1]=='W')) Write_Calibration_Data();
    else if ((cmd[0]=='W') && (cmd[1]=='+')) OverwriteBackup();
    else if ((cmd[0]=='W') && (cmd[1]=='-')) RestoreFromBackup();
    else if ((cmd[0]=='?') && (cmd[1]=='?')) HelpMe();
    else if ((cmd[0]=='!') && (cmd[1]=='!')) software_Reset();
    else if ((cmd[0]=='0') && (cmd[1]=='0')) ToggleRoundMode();
    
    else if  (cmd[0]=='C')                   SetCelsiusMode();
    else if  (cmd[0]=='F')                   SetFahrenheitdMode();
    else if  (cmd[0]=='P')                   SetReportMode();
    else if  (cmd[0]=='R')                   SetRawReadMode();
    else if  (cmd[0]=='T')                   NewReportTime();
    else if  (cmd[0]=='S')                   ToggleSensor();
     
    // ATMEGA186 does not have internal temperature senor
    else if ((cmd[0]=='I') && (cmd[1]=='T')) PrintNotImplemented();
    else if ((cmd[0]=='I') && (cmd[1]=='F')) PrintNotImplemented();
    
    // example of application specific command implimneted
    // this command writes default data to the EEPROM working storage
    else if ((cmd[0]=='Z') && (cmd[1]=='1')) TestData1();

  }

//---------------------------------------------------------------------- 
void HelpMe()
  { // We needed to seriously reduce the size of this function
    Serial.println(F(
    "; AtMega168/328 ThermometerTwo 1.0\n"
    "; ID ST RT RF FT FF F= CT CF DO RV\n"
    "; T# PF PT DB 00 L: WW W+ W- SS S#\n"
    "; LL ?? !! Z1"));
    PrintSeperatorLine();
  }
  
//----------------------------------------------------------------- 
void PrintSeperatorLine()
  { Serial.print("; ");
    for (byte i=0; i<36; i++) Serial.print('-');
    Serial.println();
  }

//================================================================  
boolean ReadTwoCharacters()
 { char c1=0,c2=0,c3=-1;
   byte m=0;
   boolean EOC=true;  // End of Command Terminator
   boolean OurReturn=false;
      
   // It is not to be believed how much effort went into creating this simple function to read 
   // two characters. I noted a bit of problem reading characters from the serail port when 
   // the loop was too fast therefore I have added a bit of a delay to insure the serial port 
   // library can keep up. Worst case senario this function can take more than 250 milliseconds.
   // Normally when this functionis called we expect the htree bytes we need to be in the buffer 
   // but if there is noise on the line or a parrot randomly pecking at the keyboard it could 
   // take a bit longer.
   // 
   // by defintion we are looking for two characters followed by a terminator
   // we define a command terminatore to be a carriage return, new line or null character
   // --- for good measure we are including the tab character and space as well
   // space was added because it is impossible to send a tab character from the Ardunion IDE
   // we will accept any combination of those characters as a single terminator
   // we will accept the last two printable ASCII characters before a terminator for our command
   // we keep reading until we get a terminator, but we will only read for a short period
   
   // but before we do anything else we are going to save the rpevious command for posterity
   prevcmd[0]=cmd[0];            // actually we are saving it so that shutdown 
   prevcmd[1]=cmd[1];            // and reset can check it before they execute
   
   while ((c3 != 13) && (c3 != 10) && (c3 != 9) && (c3 != 0) && (c3 != 32) && (m<25))
     { // if we have a valid ASCII character for c3 then roll the charaters down
       if (Serial.available()>0) 
         { c3=Serial.read();
           if (c3>32) {c1=c2; c2=c3;}
         }  
       // we need a bit of a delay to let the serial interface catch up
       // after 25 empty reads we give up
       else { delay (10); m++;}
     }
   // DebugPrintCharacters (c1,c2,c3,m);  
   
   // we are very liberal about what we will accept for a command terminator 
   // but we insist on having one.  
   if ((c3 != 13) && (c3 != 10) && (c3 != 9) && (c3 != 0) && (c3 != 32)) EOC= false;
   // we need to drain any remaining command terminator characters from the serail buffer
   else DrainCmdTermiantors();
   
   // now check for valid ASCII charaters and End of Line
   if ((c1>32) && (c2>32) && EOC)
     { // OK... we have something to work with
       // Convert lower case to UPPER case excpet "w"
       // DebugPrintCharacters (c1,c2,c3);
       if ((c1 != 'w') && (c1 >96) && (c1 <123)) c1 = (c1 -32);
       if ((c2 != 'w') && (c2 >96) && (c2 <123)) c2 = (c2 -32);
       // DebugPrintCharacters (c1,c2);
       cmd[0]=c1;
       cmd[1]=c2;
       OurReturn=true;
     }
   // whatever it was that was sent did not meet our criteria
   // inform the parrot that he or she must do better  
   else Serial.println(F("; ?? ??"));
   return OurReturn;
 }
 
//----------------------------------------------------------------
void DrainCmdTermiantors()
 { char c3=0;
   // removed leading command terminators from serial buffer 
   delay (10); c3=Serial.peek();
   while ((c3==13) || (c3==10) || (c3==9) || (c3==0) || (c3 == 32)) 
     { c3=Serial.read(); 
       delay (10); 
       c3=Serial.peek();
     }
   // c3 should at this point should be -1 unless there are more commands/charaters in the buffer
 }

//================================================================ 
// overloaded debugging function for debugging the above input routine
void DebugPrintCharacters (char c1, char c2, char c3, byte m)
  {
    if (DeBug == true)
       { Serial.print   ("Received: ");
         Serial.print   (c1);
         // Serial.print   (" ");
         Serial.print   (c2);
         Serial.print   (" ");
         if (c3 != 0) 
           { Serial.print (c3, DEC);
             Serial.print   (" ");
           }
         if (m != 0) Serial.print (m, DEC);
         Serial.println ();
       }
  } 
void DebugPrintCharacters (char c1, char c2, char c3)
  { byte m=0;
    DebugPrintCharacters (c1,c2,c3,m);
  }  
void DebugPrintCharacters (char c1, char c2)
  { byte m=0;
    char c3=0;
    DebugPrintCharacters (c1,c2,c3,m);
  } 
 
//----------------End of Main File------------------------------  

